/* ============================================================================
 Name        : huikey.c
 Author      : Nasker maig 2012
 Version     :
 Copyright   :
 Description : Objecte "HuiKey", extern per a Pure Data, adaptació del sublim [earplug~] fet pel mestre PeiXiang al 2004.
 	 	 	   Aquesta versió intenta simplificar el codi apart d'emprar les respostes impulsionals mesurades en diferents
 	 	 	   subjectes de l'IRCAM (per la seva flexibilitat), en comptes de les realitzades al MIT amb el dummy-head KEMAR.
 ============================================================================ */

#include "m_pd.h"									/* Llibreria de funcions per a l'escriptura d'externs de PureData */
#include <math.h>									/* Per a calculs aritmetics */
#include <stdio.h>									/* Lliberia standard d'entrada i sortida (per operar amb arxius) */

static t_class *huikey_class; /* Declaració de la classe, "_tilde" indica que procesa audio */

typedef struct _huikey { /* Declaració dels atributs de la classe */
	t_object x_obj; /* Defineix qualsevol objecte extern de PureData */
	t_outlet *outL; /* Definim sortida de l'objecte pel canal esquerre */
	t_outlet *outR; /* Definim sortida de l'objecte pel canal dret */

	t_float azim; /* Nombre real per definir l'azimuth d'entrada */
	t_float elev; /* Nombre real definir l'elevacio d'entrada */

	t_float sample; /* Nombre real auxiliar per operar les mostres */
	t_float buffer[128]; /* Tamany del buffer per a la convolució */
	t_float sel_hrir[2][128]; /* Resposta impulsional emprada en el moment de procés de senyal */
	t_float huikey_hrir[187][2][128]; /* Array 3D 187x2x128 per emmagatzemar les hrir's */
	t_int bufPin; /* Indica el punt d'entrada del buffer de convolució */

} t_huikey; /* Finalitzem la declaracio dels atributs de la classe */

static t_int *huikey_perform(t_int *w) { /* Implementació del mètode de procés de senyal de la classe */

	t_huikey *x = (t_huikey *) (w[1]); /* Punter a l'objecte instanciat a la classe */
	t_float *input = (t_float *) (w[2]); /* Punter a l'entrada de senyal del objecte */
	t_float *outL = (t_float *) (w[3]); /* Punter a la sortida esquerra de senyal del objecte */
	t_float *outR = (t_float *) (w[4]); /* Punter a la sortida dreta de senyal del objecte */
	int blocksize = (int) (w[5]); /* Definició del tamany del block de senyal a processar */

	int incAzim = 15; /* Per defecte l'increment d'angle és de 15º */
	int correc = 0; /* Per defecte el factor de correcció és 0 */
	int index; /* Variable on assignarem el valor de l'index dins l'array de respostes impulsionals */

	if (x->elev < -45)
		x->elev = -45; /* Controlem el rang de valors d'elevació i azimith introduits */

	if (x->elev > 90)
		x->elev = 90;

	if (x->azim < 0)
		x->azim = 0;

	if (x->azim > 360)
		x->azim = 0;

	x->elev = round(x->elev / 15) * 15; /* Quantitzem els valors a salts de 15º */

	if ((x->elev >= 60) && (x->elev < 75))
		incAzim = 30; /*Particularitzem l'increment d'angle per a valors d'elevació entre 60º i 75º */

	if ((x->elev >= 75) && (x->elev < 90)) { /* Particularitzem l'increment d'angle per a valors d'elevació entre 75º i 90º */
		incAzim = 60;
		correc = 12;
	}

	if (x->elev >= 90) { /*Particularitzem l'unic valor de sortida per a angle d'azimuth de 90º */
		x->azim = 0;
		correc = 30;
	}

	x->azim = round(x->azim / incAzim) * incAzim; /* Quantitzem el valor de l'azimuth en funció de les condicions anteriors */

	index = ((x->elev + 45) / 15) * 24 - correc + (x->azim / incAzim);
	/* Aquesta expressió calcula l'index de sortida corresponent */

	int i = 0;

	for (; i < 128; i++) { /* Aquest bucle assigna els valors de la resposta impulsional de l'array 3D al de sel_hrir */
		x->sel_hrir[0][i] = x->huikey_hrir[index][0][i];
		x->sel_hrir[1][i] = x->huikey_hrir[index][1][i];
	}

	float inSample; /* Nombre real per assignar al valor de les mostres  d'entrada */
	float sum[2]; /* Array de dos casselles on recollirem els sumatoris de la convolució */

	while (blocksize--) { /* Bucle de convolució, autodecrementa fins haver recorregut tot el block */

		sum[0] = 0; /* Inicialitzem el sumatori de la convolució a cada iteració */
		sum[1] = 0;

		inSample = *(input++); /* Incrementem el punter per triar la posició de la mostra actual a operar */

		x->buffer[x->bufPin] = inSample; /* Assignem el valor de la casella actual del buffer */

		i = 0;
		for (; i < 128; i++) { /* Iterem sobre les 128 caselles de la hrir per convolucionar */

			sum[0] += x->buffer[x->bufPin - i] * x->sel_hrir[0][i]; /* Sumatori de la convolució; assignació clau! */

			sum[1] += x->buffer[x->bufPin - i] * x->sel_hrir[1][i];

		}
		x->bufPin = (x->bufPin + 1) & 127; /* Incrementem el valor del del Punter del buffer */

		*outL++ = sum[0]; /* Treiem el valor del sumatori, canal esquerre */
		*outR++ = sum[1]; /* Treiem el valor del sumatori, canal dret */
	}
	return (w + 6); /* Retorn a l'arbre de dsp del resultat del mètode de dsp */
}

static void huikey_dsp(t_huikey *x, t_signal **sp) {

	dsp_add(huikey_perform, 5, x, sp[0]->s_vec, sp[1]->s_vec,
			sp[2]->s_vec, sp[0]->s_n);

	/* Adició al arbre de dsp de PD, els arguments indiquen: el mètode, el nombre d'arguments, entrada de senyal,
	 sortida esquerra de senyal, sortida dreta de senyal i tamany del block de mostres */

}

static void *huikey_new(t_floatarg azimArg, t_floatarg elevArg) { /* Implementació del constructor */

	t_huikey *x = (t_huikey *) pd_new(huikey_class); /* Construcció de l'objecte de la classe huikey */
	x->outL = outlet_new(&x->x_obj, gensym("signal")); /* Pin de sortida esquerre de l'objecte */
	x->outR = outlet_new(&x->x_obj, gensym("signal")); /* Pin de sortida dret de l'objecte */
	floatinlet_new(&x->x_obj, &x->azim); /* Pin d'entrada del valor d'azimuth de l'objecte */
	floatinlet_new(&x->x_obj, &x->elev); /* Pin d'entrada del valor d'elevació de l'objecte */

	x->azim = azimArg; /* Assignació de l'argument d'azimuth d'entrada a l'atribut de la classe corresponent */
	x->elev = elevArg; /* Assignació de l'argument d'elevació d'entrada a l'atribut de la classe corresponent */

	/* int i, j;	(no emprats)	/* Indexs per recorrer les caselles de l'array 3D de respostes impulsionals */
	char buff[1024], *buffptr; /* Cadena de caràcters i punter a caràcters, necessaris per mètode d'obrir arxius */
	FILE *fid; /* Punter a objecte per manipular arxius */
	int fop; /* Enter per retornar resultat de la funció d'obertura d'arxiu */

	fop = open_via_path(canvas_getdir(canvas_getcurrent())->s_name,
			"databaseHrirREDUX.txt", "", buff, &buffptr, 1024, 0);
	/* Crida a la funció de la llibreria d'externs de PD per trobar arxius a partir de la seva drecera, per defecte l'arxiu
	 * haurà de trobar-se a la mateixa carpeta que hi tinguem la llibreria dinàmica resultant de la compilació d'aquest codi
	 */

	if (fop >= 0) { /*Valors majors o iguals a 0 indiquen una que s'ha trobat l'arxiu */
		post("[huikey~] ha trobat l'arxiu amb les HRIR's!");
		post("carreguem les dades", buff); /* Indiquem al post el resultat correcte de l'operació d'obertura */
		fid = fdopen(fop, "r"); /* Crida al mètode d'obertura d'arxius */

		int i = 0;
		for (; i < 187; i++) { /* Iterem tantes vegades com respostes impulsionals volguem tinguem a l'arxiu,
		 187 per canal en aquest cas */
			while (fgetc(fid) != 10)
				; /* Realitzem l'operació mentres es pugui llegir l'arxiu; no arribem al final */
			int j = 0;
			for (; j < 128; j++) { /* Iterem pel	nombre de coeficients per resposta impulsional/cada linia */
				fscanf(fid, "%f", x->huikey_hrir[i][0][j]); /* Llegim un nombre real de l'arxiu i el situem a l'array 3D */
			}

			/* Les respostes impulsionals es troben escrites de forma entrellaçada a l'arxiu d'on les llegim */

			for (; j < 128; j++) { /* Iterem pel	nombre de coeficients per resposta impulsional/cada linia */
				fscanf(fid, "%f", x->huikey_hrir[i][1][j]); /* Llegim un nombre real de l'arxiu i el situem a l'array 3D */
			}
		}
		fclose(fid); /* Tanquem l'arxiu finalitzada la seva lectura */
	}
	return (x); /* El constructor retorna l'objecte un cop carregat */
}

void huikey_tilde_setup(void) { /* Declaració dels metodes de la classe */
	huikey_class = class_new(gensym("huikey"), /* Definició del nom de l'objecte~, l'etiqueta per la seva crida a PureData */
	(t_newmethod) huikey_new, /* Definició del seu espai de dades així com dels arguments dinàmics */
	0, sizeof(t_huikey), CLASS_DEFAULT, A_DEFFLOAT, A_DEFFLOAT, 0);

	CLASS_MAINSIGNALIN(huikey_class, t_huikey, sample);
	/* Definició del senyal d'entrada principal */

	class_addmethod(huikey_class, /* Declaració dels mètodes a afegir a l'arbre de dsp */
	(t_method) huikey_dsp, gensym("dsp"), 0);

}